# -*- coding: utf-8 -*-
"""
Created on Mon Jul  1 15:29:46 2013

@author: delande
"""
# Copyright Dominique Delande
# Provided "as is" without any warranty
#
# Computes the Floquet eigenstates of the periodically kicked rotor
# by diagonalization of the evolution operator over one period
# The latter is computed directly in momentum space, using Bessel functions
# (hopefully with the correct phase factors)
# It could also be computed using a "split" technique
# where the kick operator is computed in configuration space (where it is diagonal)
# and the free propagation is computed in momentum space (where it is diagonal)
# Passing from configuration to momentum representation is done a Fast Fourier Transform

# Because of truncation of the momentum eigenbasis, the matrix is not exactly unitary
# and some eigenvalues are not on the unit circle, but inside (worse for large K/hbar)
# This script prints the spectrum of the evolution operator truncated at -basis_size/2 and basis_size/2-1
# thus basis_size eigenvalues.
# It also prints few eigenstates (more precisely |psi[p)|**2) in the momentum basis
# States are chosen completely randomly and can thus be "unconverged" (eigenvalue of modulus
# smaller than unity) if they touch one of the boundaries.
# Execution time is dominated by the diagonalization itself, it scales as basis_size**3.
# It is roughly 10 seconds for basis_size=512
 
from math import *
import time
import socket
import random
import os
import numpy
from scipy import special
from scipy import linalg

# for generating reproducible random sequences
# comment out for irreproducible sequences  
random.seed(123)  
dpi=2.0*pi
# Execution time: 13 seconds for the following set
K=11.6
hbar=1.0
basis_size=512
n=basis_size
# Avoid quasi_momentum=0 because of discrete symmetry p <-> -p
quasi_momentum=0.1
number_of_printed_eigenstates=4
# Put random_kicked_rotor to True to generate the dynamics of the random kicked rotor
random_kicked_rotor=False

K_over_hbar=K/hbar
dpi_over_n=dpi/n
tab_u=numpy.zeros((n,n),dtype=complex)
tab_bessel=numpy.zeros(2*n)
for i in range(2*n):
  tab_bessel[i]=special.jn(i-n,K_over_hbar)
#print tab_bessel  
for i in range(n):
  if (random_kicked_rotor):
# This is for the random kicked rotor      
    phase=random.uniform(0.0,dpi)
  else:  
# and this is for the regular kicked rotor      
    phase=0.5*hbar*(i+quasi_momentum-n/2)**2-0.5*pi*i
  complex_phase_factor=complex(cos(phase),-sin(phase))
  for j in range(n):
    tab_u[j,i]=complex_phase_factor*tab_bessel[j+n-i]
    complex_phase_factor*=1j 
(w,vr)=linalg.eig(tab_u)
#print w
# output the final result
eig_file=open('Floquet_eigenvalues.dat','w')
print >>eig_file,'# Data generated on',time.asctime(),' using Python script',os.path.abspath( __file__ ),' on computer:',socket.gethostname()
print >>eig_file,'# K =',K
print >>eig_file,'# hbar =',hbar
print >>eig_file,'# basis_size =',n
print >>eig_file,'# quasi_momentum =',quasi_momentum
if (random_kicked_rotor):   
  print >>eig_file,'# Random kicked rotor'
else:
  print >>eig_file,'# Regular kicked rotor'
for i in range(n):
  print >>eig_file,w[i].real,w[i].imag
eig_file.close()

# output few eigenstates
floquet_file=open('Floquet_eigenstates.dat','w')
print >>floquet_file,'# Data generated on',time.asctime(),' using Python script',os.path.abspath( __file__ ),' on computer:',socket.gethostname()
print >>floquet_file,'# K =',K
print >>floquet_file,'# hbar =',hbar
print >>floquet_file,'# basis_size =',n
print >>floquet_file,'# quasi_momentum =',quasi_momentum
if (random_kicked_rotor):   
  print >>floquet_file,'# Random kicked rotor'
else:
  print >>floquet_file,'# Regular kicked rotor'
print >>floquet_file,'# number_of_printed_eigenstates =',number_of_printed_eigenstates
for j in range(number_of_printed_eigenstates):
  j_plot=int(n*random.uniform(0.,1.))
  print >>floquet_file,'# State',j_plot
  for i in range(n):
    print >>floquet_file,i-n/2,abs(vr[i,j_plot])**2
  print >>floquet_file,' '  
floquet_file.close()
