# -*- coding: utf-8 -*-
"""
Created on Mon Jul  1 15:29:46 2013

@author: delande
"""
# Copyright Dominique Delande
# Provided "as is" without any warranty
#
# Computes the quantum dynamics of the periodically kicked rotor
# in the presence of a constant decoherence rate
# by iteration of the evolution operator over one period.
# The latter is computed using a "split" technique
# where the kick operator is computed in configuration space (where it is diagonal)
# and the free propagation is computed in momentum space (where it is diagonal)
# Passing from configuration to momentum representation is done a Fast Fourier Transform
# For this reason, it is more efficient to use a basis size which is a power of 2
# The initial state is chosen randomly near zero momentum, in a band defined
# by the half_width_initial_state parameter
# Averaging can be performed over the initial state and (optionally) the quasi-momentum
# The expectation value <p**2> is printed at each kick
# Also the (average) density in momentum space is printed at the final time 
# Decoherence is naively modelled by a jump in the quasi-momentum
# (like in spontaneous emission)
 
from math import *
import time
import socket
import random
import os
import numpy
from scipy import fftpack

def compute_p2(wave_function,tab_momentum):
#  n=wave_function.size
#  p2=0.0  
#  norm=0.0
#  for i in range(n):
#    p2+=((i+quasi_momentum-n/2)*abs(wave_function[i]))**2
#    norm+=abs(wave_function[i])**2
  p2=numpy.sum((abs(wave_function)*tab_momentum)**2)
  norm=numpy.sum(abs(wave_function)**2)
#  norm=1.0
#  print norm,norm2 
  return p2/norm

# for generating reproducible random sequences
# comment out for irreproducible sequences  
random.seed(123)  
dpi=2.0*pi
K=11.6
# Execution time: 28 seconds for the following set
hbar=1.0
number_of_initial_states=100
basis_size=1024
final_time=1000
decoherence_rate=0.005

#
n=basis_size
# put half_width_initial_state to 0 for no averaging over initial state
half_width_initial_state=2


K_over_hbar=K/hbar
dpi_over_n=dpi/n
tab_p2=numpy.zeros(final_time+1)
tab_density=numpy.zeros(n)
tab_momentum=numpy.zeros(n)
tab_spatial_phase=numpy.zeros(n)
tab_local_phase=numpy.zeros(n)
tab_kick_phase_factor=numpy.zeros(n,dtype=complex)
wave_function=numpy.zeros(basis_size,dtype=complex)
tab_free_prop_phase_factor=numpy.zeros(basis_size,dtype=complex)
for i in range(n):
  tab_spatial_phase[i]=cos(i*dpi_over_n)
  tab_kick_phase_factor[i]=complex(cos(K_over_hbar*tab_spatial_phase[i]),-sin(K_over_hbar*tab_spatial_phase[i]))
for i_state in range(number_of_initial_states):
# prepare initial state
# choose the quasi-momentum and populate only few momentum states near p=0 with random phases
  quasi_momentum=random.uniform(-0.5,0.5)
#  print 'Now starts initial state',i_state,quasi_momentum  
  for i in range(n):
    tab_momentum[i]=i+quasi_momentum-n/2 
    phase=0.5*hbar*(i+quasi_momentum-n/2)**2
    tab_free_prop_phase_factor[i]=complex(cos(phase),-sin(phase))
  wave_function[0:n]=0.0
  for i in range(n/2-half_width_initial_state,n/2+half_width_initial_state+1):
    phase=random.uniform(0,dpi)
    wave_function[i]=complex(cos(phase),sin(phase))/sqrt(2*half_width_initial_state+1)
#  tab_p2[0]+=compute_p2(wave_function,tab_momentum)
  tab_p2[0]+=numpy.sum((numpy.real(wave_function)*tab_momentum)**2+(numpy.imag(wave_function)*tab_momentum)**2)
# now start time propagation
  for my_time in range(1,final_time+1):  
# first free propagation
    wave_function*=tab_free_prop_phase_factor
# then apply the kick
# this requires first to convert to configuration space by a FFT
    wave_function=fftpack.ifft(wave_function,overwrite_x=True)
# then multiply by the kick operator
    wave_function*=tab_kick_phase_factor   
# back in momentum space by a FFT      
    wave_function=fftpack.fft(wave_function,overwrite_x=True)
#    tab_p2[my_time]+=compute_p2(wave_function,tab_momentum)
    tab_p2[my_time]+=numpy.sum((numpy.real(wave_function)*tab_momentum)**2+(numpy.imag(wave_function)*tab_momentum)**2)
# Is there a decoherence event?
    if (random.uniform(0.,1.)<decoherence_rate):
# yes! I reset the quasi-momentum to a random value    
      quasi_momentum=random.uniform(-0.5,0.5)
      for i in range(n):
        tab_momentum[i]=i+quasi_momentum-n/2 
        phase=0.5*hbar*(i+quasi_momentum-n/2)**2
        tab_free_prop_phase_factor[i]=complex(cos(phase),-sin(phase))     
# save the final (squared) wavefunction 
  tab_density+=numpy.real(wave_function)**2+numpy.imag(wave_function)**2
# output the final result
p2_file=open('p2_vs_time_quantum_with_fixed_decoherence_rate.dat','w')
print >>p2_file,'# Data generated on',time.asctime(),' using Python script',os.path.abspath( __file__ ),' on computer:',socket.gethostname()
print >>p2_file,'# Regular kicked rotor'
print >>p2_file,'# With averaging over quasi-momentum'     
print >>p2_file,'# K =',K
print >>p2_file,'# number_of_initial_states =',number_of_initial_states
print >>p2_file,'# half_width_initial_state =',half_width_initial_state
print >>p2_file,'# final_time =',final_time
print >>p2_file,'# basis_size =',n
for my_time in range(final_time+1):  
  print >>p2_file,my_time,hbar**2*tab_p2[my_time]/number_of_initial_states
p2_file.close()
density_file=open('density_momentum_with_fixed_decoherence_rate.dat','w')
print >>density_file,'# Data generated on',time.asctime(),' using Python script',os.path.abspath( __file__ ),' on computer:',socket.gethostname()
print >>density_file,'# Regular kicked rotor'
print >>density_file,'# With averaging over quasi-momentum'     
print >>density_file,'# K =',K
print >>density_file,'# number_of_initial_states =',number_of_initial_states
print >>density_file,'# half_width_initial_state =',half_width_initial_state
print >>density_file,'# final_time =',final_time
print >>density_file,'# basis_size =',n
for i in range(n):  
  print >>density_file,i-n/2,tab_density[i]/number_of_initial_states
density_file.close()
